let opt_tbl_sz = 17
let builtins_tbl_sz = 17

open Util

open Format

open Kernel

module Builtin = struct
  type desc = {
    s: Sgn.t;
  }

  module Void = struct
    type spec =
      | Simple
      | Do

    type t = {
      desc: desc;
      spec: spec;
    }
  end

  module Value = struct
    type t = {
      desc: desc;
    }
  end

  module Bool = struct
    type t = {
      desc: desc;
    }
  end

  type t =
    | Void of Void.t
    | Value of Value.t
    | Bool of Bool.t
end

let builtins_tbl = Sgn.Hashtbl.create builtins_tbl_sz

let subst_var s x =
  try Value.to_var Pos.dummy_range (Var.Map.find x s)
  with Not_found -> x

let enum_vars_in_value set v =
  Value.fold_vars Var.Set.add v set

let enum_vars_in_compound set c =
  Compound.fold_left enum_vars_in_value set c

module Call = struct
  module Void = struct
    type t = {
      builtin: Builtin.Void.t;
      args: Value.t array;
    }

    let format fmt c =
      fprintf fmt "call %s(%a)"
        c.builtin.Builtin.Void.desc.Builtin.s.Sgn.f
        (Array.format "," Value.format) c.args

    let subst s c = {
      builtin = c.builtin;
      args = Array.map (Value.subst s) c.args;
    }

    let enum_vars set c = Array.fold_left enum_vars_in_value set c.args
  end

  module Value = struct
    type t = {
      builtin: Builtin.Value.t;
      args: Value.t array;
      result: Var.t;
    }

    let format fmt c =
      fprintf fmt "%a <- call %s(%a)"
        Var.format c.result
        c.builtin.Builtin.Value.desc.Builtin.s.Sgn.f
        (Array.format "," Value.format) c.args

    let subst s c = {
      builtin = c.builtin;
      args = Array.map (Value.subst s) c.args;
      result = subst_var s c.result;
    }

    let enum_vars set c =
      Var.Set.add c.result (Array.fold_left enum_vars_in_value set c.args)
  end

  type t =
    | Void of Void.t
    | Value of Value.t

  let format fmt c =
    match c with
      | Void c -> Void.format fmt c
      | Value c -> Value.format fmt c

  let subst s c =
    match c with
      | Void c -> Void (Void.subst s c)
      | Value c -> Value (Value.subst s c)

  let enum_vars set c =
    match c with
      | Void c -> Void.enum_vars set c
      | Value c -> Value.enum_vars set c
end

type t = {
  vars: Var.Set.t;
  tells: (Pos.range * Value.t array) list Sgn.Map.t Var.Map.t;
  calls: (Pos.range * Call.t) list;
  asks: (Pos.range * Var.t array * t) list Sgn.Map.t Var.Map.t;
  bool_calls: (Pos.range * Sgn.t * Value.t array * t * t) list;
}

let is_empty opt =
  Var.Set.is_empty opt.vars &&
  Var.Map.is_empty opt.tells &&
  opt.calls = [] &&
  Var.Map.is_empty opt.asks &&
  opt.bool_calls = []

let map_iter f map =
  let sgn_iter x map = Sgn.Map.iter (fun s list -> List.iter (f x s) list) map in
  Var.Map.iter sgn_iter map

let map_fold f v map =
  let sgn_fold x map v =
    Sgn.Map.fold (fun s list v -> List.fold_left (f x s) v list) map v in
  Var.Map.fold sgn_fold map v

let map_map f map =
  Var.Map.map (Sgn.Map.map (List.map f)) map

let rec format fmt opt =
  if not (Var.Set.is_empty opt.vars) then
    begin
      fprintf fmt "exists %a" (Var.Set.format " " Var.format) opt.vars;
      fprintf fmt "{"
    end;
  let first = ref true in
  let check_first () =
    if !first then
        first := false
    else
      fprintf fmt " || " in
  let print_call (_, c) =
    check_first ();
    Call.format fmt c in
  List.iter print_call opt.calls;
  let print_tell v s (_, a) =
    check_first ();
    fprintf fmt "%a.%s(%a)" Var.format v s.Sgn.f
      (Array.format "," Value.format) a in
  map_iter print_tell opt.tells;
  let print_ask v s (_, a, r) =
    check_first ();
    fprintf fmt "%a.%s(%a){%a}"
      Var.format v s.Sgn.f (Array.format "," Var.format) a format r in
  map_iter print_ask opt.asks;
  let print_bool_call (_, s, a, t, f) =
    check_first ();
    fprintf fmt "if %s(%a) {%a} else {%a}"
      s.Sgn.f (Array.format "," Value.format) a format t format f in
  List.iter print_bool_call opt.bool_calls;
  if !first then
    fprintf fmt "1";
  if not (Var.Set.is_empty opt.vars) then
    fprintf fmt "}"

let empty = {
  vars = Var.Set.empty;
  tells = Var.Map.empty;
  calls = [];
  asks = Var.Map.empty;
  bool_calls = [];
}

let add_fresh x (s, vars) =
  let y = Var.fresh () in
  Var.Map.add x (Value.Var y) s, Var.Set.add y vars

let subst_tell s (pos, a) =
  (pos, Array.map (Value.subst s) a)

let subst_tell_sgn sub s list tells =
  let list = List.map (subst_tell sub) list in
  Sgn.Map.add s list tells

let subst_tell_var s x sgn_map tells =
  let sgn_map = Sgn.Map.fold (subst_tell_sgn s) sgn_map Sgn.Map.empty in
  Var.Map.add (subst_var s x) sgn_map tells

let subst_ask subst s (pos, a, opt) =
  let s = ref s in
  let make_subst x =
    let y = Var.fresh () in
    s := Var.Map.add x (Value.Var y) !s;
    y in
  let a = Array.map make_subst a in
  (pos, a, subst !s opt)

let subst_ask_var subst s x sgn_map asks =
  let sgn_map = Sgn.Map.map (List.map (subst_ask subst s)) sgn_map in
  Var.Map.add (subst_var s x) sgn_map asks

let subst_bool_call subst sub (pos, s, c, t, f) =
  (pos, s, Array.map (Value.subst sub) c, subst sub t, subst sub f)

let rec subst s t =
  let (s, vars) = Var.Set.fold add_fresh t.vars (s, Var.Set.empty) in
  let tells = Var.Map.fold (subst_tell_var s) t.tells Var.Map.empty in
  let calls = List.map (fun (pos, c) -> (pos, Call.subst s c)) t.calls in
  let asks = Var.Map.fold (subst_ask_var subst s) t.asks Var.Map.empty in
  let bool_calls = List.map (subst_bool_call subst s) t.bool_calls in
  { vars = vars; tells = tells; calls = calls; asks = asks;
    bool_calls = bool_calls }

let var_map_add x map maps =
  if Sgn.Map.is_empty map then Var.Map.remove x maps
  else Var.Map.add x map maps

let sgn_list_add s list maps =
  if list = [] then Sgn.Map.remove s maps
  else Sgn.Map.add s list maps

let var_map_get x map =
  try Var.Map.find x map
  with Not_found -> Sgn.Map.empty

let sgn_list_get s map =
  try Sgn.Map.find s map
  with Not_found -> []

let union_sgn_map s list sgn_map' =
  let list =
    try
      let list' = Sgn.Map.find s sgn_map' in
      List.rev_append list list'
    with Not_found -> list in
  Sgn.Map.add s list sgn_map'

let union_var_map x sgn_map var_map' =
  let sgn_map =
    try
      let sgn_map' = Var.Map.find x var_map' in
      Sgn.Map.fold union_sgn_map sgn_map sgn_map'
    with Not_found -> sgn_map in
  Var.Map.add x sgn_map var_map'

let union r r' =
  assert (Var.Set.is_empty (Var.Set.inter r.vars r'.vars));
  {
    vars = Var.Set.union r.vars r'.vars;
    tells = Var.Map.fold union_var_map r.tells r'.tells;
    calls = r.calls @ r'.calls;
    asks = Var.Map.fold union_var_map r.asks r'.asks;
    bool_calls = r.bool_calls @ r'.bool_calls;
  }

let opt_tbl: (t -> t) String.Hashtbl.t = String.Hashtbl.create opt_tbl_sz

let show_help fmt =
  let show_opt s _ =
    Format.fprintf fmt "  %s@." s in
  String.Hashtbl.iter show_opt opt_tbl

let add_var var opt =
  { opt with vars = Var.Set.add var opt.vars }

let 
singleton_map v s list =
  Var.Map.add v (Sgn.Map.add s list Sgn.Map.empty) Var.Map.empty

let make_tell pos v f args =
  let s = Sgn.get f (Array.length args) in
  singleton_map v s [(pos, args)]

let of_void_call pos b c =
  let (args, tells) =
    match b.Builtin.Void.spec with
      | Builtin.Void.Simple ->
          (c.Compound.args, Var.Map.empty)
      | Builtin.Void.Do ->
          let args = c.Compound.args in
          let args_count = Array.length args in
          let r = Value.to_var pos (args.(args_count - 1)) in
          let args = Array.sub args 0 (args_count - 1) in
          (args, make_tell pos r "done" [| |]) in
  let c = {
    Call.Void.builtin = b;
    args = args
  } in
  { empty with
      calls = [(pos, Call.Void c)];
      tells = tells }

let of_value_call pos b c =
  let result = Var.fresh () in
  let args = c.Compound.args in
  let args_count = Array.length args in
  let r = Value.to_var pos (args.(args_count - 1)) in
  let args = Array.sub args 0 (args_count - 1) in
  let c = {
    Call.Value.builtin = b;
    args = args;
    result = result;
  } in
  { empty with
      vars = Var.Set.singleton result;
      calls = [(pos, Call.Value c)];
      tells = make_tell pos r "value" [| Value.Var result |];
  }

let of_bool_call pos b c =
  let args = c.Compound.args in
  let args_count = Array.length args in
  let r = Value.to_var pos (args.(args_count - 1)) in
  let args = Array.sub args 0 (args_count - 1) in
  let tell_true = { empty with tells = make_tell pos r "true" [| |] } in
  let tell_false = { empty with tells = make_tell pos r "false" [| |] } in
  { empty with bool_calls = [(pos, c.Compound.s, args, tell_true, tell_false)] }

let of_call pos c =
  try
    match Sgn.Hashtbl.find builtins_tbl c.Compound.s with
      | Builtin.Void b -> of_void_call pos b c
      | Builtin.Value b -> of_value_call pos b c
      | Builtin.Bool b -> of_bool_call pos b c
  with Not_found ->
    Err.unknown_builtin pos (Compound.f c) (Compound.arity c)

let rec of_agent a =
  match a with
    | Agent.Skip -> empty
    | Agent.Tell (pos, _, v, c) ->
        let list = [(pos, c.Compound.args)] in
        let sgns = Sgn.Map.add c.Compound.s list Sgn.Map.empty in
        let tells = Var.Map.add v sgns Var.Map.empty in
        { empty with tells = tells }
    | Agent.Call (pos, c) -> of_call pos c
    | Agent.Ask (pos, v, c, a) ->
        let list = [(pos, c.Compound.args, of_agent a)] in
        { empty with asks = singleton_map v c.Compound.s list }
    | Agent.Exists (v, a) ->
        let r = of_agent a in
        { r with vars = Var.Set.add v r.vars }
    | Agent.Par (a0, a1) ->
        union (of_agent a0) (of_agent a1)

type usage = {
  bound_vars: Var.Set.t;
  non_local_vars: Var.Set.t;
  occur_asks: Sgn.Set.t Var.Map.t;
  occur_tells: Sgn.Set.t Var.Map.t;
  pending_codes: t list;
  unknown_asks: t list Sgn.Map.t Var.Map.t;
  unknown_tells: Var.Set.t Sgn.Map.t Var.Map.t;
}

let empty_usage = {
  bound_vars = Var.Set.empty;
  non_local_vars = Var.Set.empty;
  occur_asks = Var.Map.empty;
  occur_tells = Var.Map.empty;
  pending_codes = [];
  unknown_asks = Var.Map.empty;
  unknown_tells = Var.Map.empty;
}

let add_ask_to_pending_code pending_codes (_, _, opt) = opt :: pending_codes

let rec enum_declare_non_local_var x usage =
  if Var.Set.mem x usage.non_local_vars then
    usage
  else
    let usage =
      { usage with non_local_vars = Var.Set.add x usage.non_local_vars } in
    let usage =
      try
        let unknown_tells' = Var.Map.find x usage.unknown_tells in
        let usage =
          { usage with
              unknown_tells = Var.Map.remove x usage.unknown_tells } in
        let new_vars =
          Sgn.Map.fold (fun _ set set' -> Var.Set.union set set') unknown_tells'
            Var.Set.empty in
        Var.Set.fold enum_declare_non_local_var new_vars usage
      with Not_found -> usage in
    let usage =
      try
        let unknown_asks' = Var.Map.find x usage.unknown_asks in
        let unknown_asks = Var.Map.remove x usage.unknown_asks in
        let pending_codes =
          Sgn.Map.fold
            (fun _ list pending_codes -> List.rev_append list pending_codes)
            unknown_asks' usage.pending_codes in
        { usage with
            pending_codes = pending_codes;
            unknown_asks = unknown_asks }
      with Not_found -> usage in
    usage
    
let enum_usage_non_local_asks s list pending_codes =
  List.fold_left add_ask_to_pending_code pending_codes list

let enum_usage_local_asks occur_tells s list
    (occur_asks, pending_codes, unknown_asks, unknown_tells, pending_vars) =
  let occur_asks = Sgn.Set.add s occur_asks in
  if Sgn.Set.mem s occur_tells then
    let pending_codes =
      List.fold_left add_ask_to_pending_code pending_codes list in
    let unknown_tells, pending_vars =
      try
        let new_vars = Sgn.Map.find s unknown_tells in
        let unknown_tells = Sgn.Map.remove s unknown_tells in
        let pending_vars = Var.Set.union pending_vars new_vars in
        unknown_tells, pending_vars
      with Not_found -> unknown_tells, pending_vars in
    (occur_asks, pending_codes, unknown_asks, unknown_tells, pending_vars)
  else
    let list = List.rev_map (fun (_, _, opt) -> opt) list in
    let list' = sgn_list_get s unknown_asks in
    let list = List.rev_append list list' in
    (occur_asks, pending_codes, Sgn.Map.add s list unknown_asks,
     unknown_tells, pending_vars)

let set_of_var_array vars a =
  Array.fold_right Var.Set.add a vars

let enum_non_local_vars_in_asks _ list non_local_vars =
  List.fold_left (fun vars (_, a, _) -> set_of_var_array vars a) non_local_vars
    list

let enum_usage_asks x sgn_map usage =
  let non_local_vars =
    Sgn.Map.fold enum_non_local_vars_in_asks sgn_map usage.non_local_vars in
  let usage = { usage with non_local_vars = non_local_vars } in
  if Var.Set.mem x usage.non_local_vars
    || not (Var.Set.mem x usage.bound_vars) then
    let pending_codes =
      Sgn.Map.fold enum_usage_non_local_asks sgn_map usage.pending_codes in
    { usage with pending_codes = pending_codes }
  else
    let occur_asks =
      try Var.Map.find x usage.occur_asks
      with Not_found -> Sgn.Set.empty in
    let occur_tells =
      try Var.Map.find x usage.occur_tells
      with Not_found -> Sgn.Set.empty in
    let unknown_asks = var_map_get x usage.unknown_asks in
    let unknown_tells = var_map_get x usage.unknown_tells in
    let (occur_asks, pending_codes, unknown_asks, unknown_tells,
         pending_vars) =
      Sgn.Map.fold (enum_usage_local_asks occur_tells) sgn_map
        (occur_asks, usage.pending_codes, unknown_asks, unknown_tells,
         Var.Set.empty) in
    let usage =
      { usage with
          occur_asks = Var.Map.add x occur_asks usage.occur_asks;
          pending_codes = pending_codes;
          unknown_asks = Var.Map.add x unknown_asks usage.unknown_asks;
          unknown_tells = Var.Map.add x unknown_tells usage.unknown_tells;
      } in
    Var.Set.fold enum_declare_non_local_var pending_vars usage

let enum_vars_in_tell_list vars list =
  let enum_array vars (_, a) =
    Array.fold_left enum_vars_in_value vars a in
  List.fold_left enum_array vars list

let enum_vars_in_tell_sgn_map vars sgn_map =
  let add_tells _ list vars =
    enum_vars_in_tell_list vars list in
  Sgn.Map.fold add_tells sgn_map vars

let enum_unsafe_vars_in_tell_var_map vars var_map =
  let add_tells _ map vars =
    enum_vars_in_tell_sgn_map vars map in
  Var.Map.fold add_tells var_map vars

let enum_non_local_vars_in_tells occur_asks s list (vars, unknown_tells) =
  let set = enum_vars_in_tell_list Var.Set.empty list in
  if Sgn.Set.mem s occur_asks then
    (Var.Set.union set vars, unknown_tells)
  else
    let set =
      try Var.Set.union set (Sgn.Map.find s unknown_tells)
      with Not_found -> set in
    (vars, Sgn.Map.add s set unknown_tells)

let enum_usage_local_tells s _ (occur_tells, pending_codes, unknown_asks) =
  if Sgn.Set.mem s occur_tells then
    (occur_tells, pending_codes, unknown_asks)
  else
    let occur_tells = Sgn.Set.add s occur_tells in
    let (pending_codes, unknown_asks) =
      match sgn_list_get s unknown_asks with
        | [] -> (pending_codes, unknown_asks)
        | non_empty_list ->
            (List.rev_append non_empty_list pending_codes,
             Sgn.Map.remove s unknown_asks) in
    (occur_tells, pending_codes, unknown_asks)

let enum_usage_tells x sgn_map usage =
  if Var.Set.mem x usage.non_local_vars
    || not (Var.Set.mem x usage.bound_vars) then
      let vars = enum_vars_in_tell_sgn_map Var.Set.empty sgn_map in
      Var.Set.fold enum_declare_non_local_var vars usage
  else
    let occur_asks =
      try Var.Map.find x usage.occur_asks
      with Not_found -> Sgn.Set.empty in
    let unknown_tells = var_map_get x usage.unknown_tells in
    let (vars, unknown_tells) =
      Sgn.Map.fold (enum_non_local_vars_in_tells occur_asks) sgn_map
        (Var.Set.empty, unknown_tells) in
    let usage = {
      usage with
        unknown_tells = var_map_add x unknown_tells usage.unknown_tells } in
    let usage = Var.Set.fold enum_declare_non_local_var vars usage in
    let occur_tells =
      try Var.Map.find x usage.occur_tells
      with Not_found -> Sgn.Set.empty in
    let unknown_asks = var_map_get x usage.unknown_asks in
    let (occur_tells, pending_codes, unknown_asks) =
      Sgn.Map.fold enum_usage_local_tells sgn_map
        (occur_tells, usage.pending_codes, unknown_asks) in
    { usage with
        occur_tells = Var.Map.add x occur_tells usage.occur_tells;
        pending_codes = pending_codes;
        unknown_asks = var_map_add x unknown_asks usage.unknown_asks;
    }

let enum_usage_calls usage (_, c) =
  Var.Set.fold enum_declare_non_local_var (Call.enum_vars Var.Set.empty c) usage

let enum_usage_bool_calls usage (_, _, a, t, f) =
  let vars = Array.fold_left enum_vars_in_value Var.Set.empty a in
  let usage = Var.Set.fold enum_declare_non_local_var vars usage in
  { usage with pending_codes = t :: f :: usage.pending_codes }

let rec enum_usage usage opt =
  let usage =
    { usage with bound_vars = Var.Set.union opt.vars usage.bound_vars } in
  let usage = Var.Map.fold enum_usage_asks opt.asks usage in
  let usage = Var.Map.fold enum_usage_tells opt.tells usage in
  let usage = List.fold_left enum_usage_calls usage opt.calls in
  let usage = List.fold_left enum_usage_bool_calls usage opt.bool_calls in
  let pending_codes = usage.pending_codes in
  let usage = { usage with pending_codes = [] } in
  List.fold_left enum_usage usage pending_codes

let removed = ref false

let remove_unused_in_tells' asks s list (used_vars, tells) =
  if Sgn.Set.mem s asks then
    let used_vars = enum_vars_in_tell_list used_vars list in
    let tells = Sgn.Map.add s list tells in
    (used_vars, tells)
  else
    begin
      removed := true;
      (used_vars, tells)
    end

let remove_unused_in_tells usage x sgn_map (used_vars, tells) =
  if Var.Set.mem x usage.non_local_vars
    || not (Var.Set.mem x usage.bound_vars) then
    let add_tell_var _ list used_vars =
      enum_vars_in_tell_list used_vars list in
    let used_vars = Sgn.Map.fold add_tell_var sgn_map used_vars in
    let tells = Var.Map.add x sgn_map tells in
    (Var.Set.add x used_vars, tells)
  else
    try
      let asks = Var.Map.find x usage.occur_asks in
      let (used_vars, tells') =
        Sgn.Map.fold (remove_unused_in_tells' asks) sgn_map
          (used_vars, Sgn.Map.empty) in
      if Sgn.Map.is_empty tells' then
        (used_vars, tells)
      else
        (Var.Set.add x used_vars, Var.Map.add x tells' tells)
    with Not_found ->
      removed := true;
      (used_vars, tells)

let remove_unused_in_rec remove_unused_in (used_vars, list) (pos, a, opt) =
  let (used_vars', opt) = remove_unused_in opt in
  if is_empty opt then
    begin
      removed := true;
      (used_vars, list)
    end
  else
    (Var.Set.union used_vars used_vars', (pos, a, opt) :: list)

let remove_unused_in_asks' remove_unused_in tells s list (used_vars, asks) =
  if Sgn.Set.mem s tells then
    let (used_vars, list) =
      List.fold_left (remove_unused_in_rec remove_unused_in) (used_vars, [])
        list in
    let asks = sgn_list_add s list asks in
    (used_vars, asks)
  else
    begin
      removed := true;
      (used_vars, asks)
    end

let remove_unused_in_asks remove_unused_in usage x sgn_map (used_vars, asks) =
  if Var.Set.mem x usage.non_local_vars
    || not (Var.Set.mem x usage.bound_vars) then
    let add_ask s list (used_vars, asks) =
      let (used_vars, list) =
        List.fold_left (remove_unused_in_rec remove_unused_in) (used_vars, [])
          list in
      let asks = sgn_list_add s list asks in
      (used_vars, asks) in
    let (used_vars, asks') =
      Sgn.Map.fold add_ask sgn_map (used_vars, Sgn.Map.empty) in
    (Var.Set.add x used_vars, var_map_add x asks' asks)
  else
    try
      let tells = Var.Map.find x usage.occur_tells in
      let (used_vars, asks') =
        Sgn.Map.fold (remove_unused_in_asks' remove_unused_in tells) sgn_map
          (used_vars, Sgn.Map.empty) in
      if Sgn.Map.is_empty asks' then
        (used_vars, asks)
      else
        (Var.Set.add x used_vars, Var.Map.add x asks' asks)
    with Not_found ->
      removed := true;
      (used_vars, asks)

let remove_unused_in_bool_calls remove_unused_in usage (used_vars, bool_calls)
    (pos, s, a, t, f) =
  let used_vars = Array.fold_left enum_vars_in_value used_vars a in
  let used_vars_t, t = remove_unused_in t in
  let used_vars_f, f = remove_unused_in f in
  let used_vars' = Var.Set.union used_vars_t used_vars_f in
  if is_empty t && is_empty f then
    begin
      removed := true;
      (used_vars, bool_calls)
    end
  else
    (Var.Set.union used_vars used_vars', (pos, s, a, t, f) :: bool_calls)
  
let rec remove_unused_in usage opt =
  let (used_vars, tells) =
    Var.Map.fold (remove_unused_in_tells usage) opt.tells
      (Var.Set.empty, Var.Map.empty) in
  let (used_vars, asks) =
    Var.Map.fold (remove_unused_in_asks (remove_unused_in usage) usage) opt.asks
      (used_vars, Var.Map.empty) in
  let (used_vars, bool_calls) =
    List.fold_left (remove_unused_in_bool_calls (remove_unused_in usage) usage)
      (used_vars, []) opt.bool_calls in
  let used_vars =
    List.fold_left
      (fun used_vars (_, c) -> Call.enum_vars used_vars c)
      used_vars opt.calls in
  let vars = Var.Set.inter opt.vars used_vars in
  let opt = {
    opt with
      asks = asks; tells = tells; vars = vars; bool_calls = bool_calls } in
  used_vars, opt

let remove_unused' opt =
  let usage = enum_usage empty_usage opt in
  let used_vars, opt = remove_unused_in usage opt in
  ignore used_vars;
  opt

let rec remove_unused opt =
  removed := false;
  let opt = remove_unused' opt in
  if !removed then
    remove_unused opt
  else
    opt

let () = String.Hashtbl.add opt_tbl "remove-unused" remove_unused

let inlined = ref false

let enum_vars_in_call vars (_, c) = Call.enum_vars vars c

let rec enum_unsafe_variables' vars opt =
  let vars = enum_unsafe_vars_in_tell_var_map vars opt.tells in
  let vars = List.fold_left enum_vars_in_call vars opt.calls in
  let enum_bool_call vars (_, _, a, t, f) =
    let vars = Array.fold_left enum_vars_in_value vars a in
    enum_unsafe_variables' (enum_unsafe_variables' vars t) f in
  let vars = List.fold_left enum_bool_call vars opt.bool_calls in
  let enum_ask _ _ vars (_, _, opt) = enum_unsafe_variables' vars opt in
  map_fold enum_ask vars opt.asks

let enum_unsafe_variables = enum_unsafe_variables' Var.Set.empty

let rec count_tell_rec counter x s opt =
  begin
    try
      let sgn_map = Var.Map.find x opt.tells in
      let list = Sgn.Map.find s sgn_map in
      counter := !counter + List.length list
    with Not_found -> ()
  end;
  let count_bool_call (_, _, _, t, f) =
    count_tell_rec counter x s t;
    count_tell_rec counter x s f in
  List.iter count_bool_call opt.bool_calls;
  map_iter (fun _ _ (_, _, opt) -> count_tell_rec counter x s opt) opt.asks

let count_tell x s opt =
  let counter = ref 0 in
  count_tell_rec counter x s opt;
  !counter

let rec no_sub_asks_rec x s opt =
  map_iter (fun _ _ (_, _, opt) -> no_asks x s opt) opt.asks;
  List.iter (fun (_, _, _, t, f) -> no_asks x s t; no_asks x s f) opt.bool_calls
and no_asks x s opt =
  try
    let sgn_map = Var.Map.find x opt.asks in
    let _ = Sgn.Map.find s sgn_map in
    raise Exit
  with Not_found ->
    no_sub_asks_rec x s opt

let no_sub_asks x s opt =
  try
    no_sub_asks_rec x s opt;
    true
  with Exit -> false

let small_ask (_, _, opt) =
  Var.Map.is_empty opt.asks

(*
  match list with
    | [single]
        when (small_ask single || only_one_tell x s opt) && no_sub_asks x s opt ->
        (Sgn.Map.add s list safe_asks, other_asks)
    | _ ->
*)

let enum_inlinable_asks criterion unsafe_variables opt x sgn_map
    inlinable_asks =
  if Var.Set.mem x opt.vars && not (Var.Set.mem x unsafe_variables) then
    let filter s list inlinable_asks =
      match criterion opt x s list with
        | Some context_criterion when no_sub_asks x s opt ->
            Sgn.Map.add s (context_criterion, list) inlinable_asks
        | _ ->
            inlinable_asks in
    let inlinable_asks' = Sgn.Map.fold filter sgn_map Sgn.Map.empty in
    var_map_add x inlinable_asks' inlinable_asks
  else
    inlinable_asks

let make_subst a a' =
  let rec aux s i =
    if i < Array.length a then
      aux (Var.Map.add a.(i) a'.(i) s) (succ i)
    else
      s in
  aux Var.Map.empty 0

let add_inline context_criterion depth ask_list (list_to_inline, left)
    (pos, a) =
  if context_criterion depth a then
    let opt =
      match ask_list with
        | [(pos, a', opt)] ->
            subst (make_subst a' a) opt
        | _ ->
            let choice = Var.fresh () in
            let make_choice (pos, a', opt) =
              (pos, [| |], subst (make_subst a' a) opt) in
            let choices = List.map make_choice ask_list in
            { empty with
                vars = Var.Set.singleton choice;
                tells = make_tell pos choice "choice" [| |];
                asks = singleton_map choice (Sgn.get "choice" 0) choices;
            } in
    (opt :: list_to_inline, left)
  else
    (list_to_inline, (pos, a) :: left)

let inline_asks_sgn depth inlinable_asks s list (list_to_inline, tells) =
  try
    let (context_criterion, ask_list) = Sgn.Map.find s inlinable_asks in
    let (list_to_inline, left) =
      List.fold_left
        (add_inline context_criterion depth ask_list)
        (list_to_inline, []) list in
    inlined := true;
    let tells =
      if left = [] then tells
      else Sgn.Map.add s left tells in
    (list_to_inline, tells)
  with Not_found -> (list_to_inline, Sgn.Map.add s list tells)

let inline_asks_var depth inlinable_asks x sgn_map (list_to_inline, tells) =
  let (list_to_inline, sgn_map) =
    try
      let sgn_map' = Var.Map.find x inlinable_asks in
      Sgn.Map.fold (inline_asks_sgn depth sgn_map') sgn_map
        (list_to_inline, Sgn.Map.empty)
    with Not_found ->
      (list_to_inline, sgn_map) in
  (list_to_inline, Var.Map.add x sgn_map tells)

let sub_apply f opt =
  let sub_apply_ask (pos, a, opt) = (pos, a, f opt) in
  let asks = map_map sub_apply_ask opt.asks in
  let sub_apply_bool_call (pos, s, a, case_true, case_false) =
    (pos, s, a, f case_true, f case_false) in
  let bool_calls = List.map sub_apply_bool_call opt.bool_calls in
  { opt with asks = asks; bool_calls = bool_calls }

let rec inline_asks depth inlinable_asks opt =
  let opt = sub_apply (inline_asks (succ depth) inlinable_asks) opt in
  let (list_to_inline, other_tells) =
    Var.Map.fold (inline_asks_var depth inlinable_asks) opt.tells
      ([], Var.Map.empty) in
  let opt = { opt with tells = other_tells } in
  List.fold_left union opt list_to_inline

let rec inlining' criterion unsafe_variables opt =
  let opt = sub_apply (inlining' criterion unsafe_variables) opt in
  let inlinable_asks =
    Var.Map.fold (enum_inlinable_asks criterion unsafe_variables opt) opt.asks
      Var.Map.empty in
  inline_asks 0 inlinable_asks opt

let inlining_once criterion opt =
  let opt = remove_unused opt in
  let unsafe_variables = enum_unsafe_variables opt in
  inlining' criterion unsafe_variables opt

let rec inlining_until_fixpoint criterion opt =
  inlined := false;
  let opt = inlining_once criterion opt in
  if !inlined then
    inlining_until_fixpoint criterion opt
  else
    opt

let safe_inlining_criterion opt x s list =
  let b =
    match list with
      | [single] -> small_ask single || count_tell x s opt = 1
      | _ -> s.Sgn.arity >= 1 && count_tell x s opt = 1 in
  if b then
    if List.for_all (fun (_, _, opt) -> count_tell x s opt = 0) list then
      Some (fun _ _ -> true)
    else
      None
  else
    None

let safe_inlining_once = inlining_once safe_inlining_criterion

let () = String.Hashtbl.add opt_tbl "safe-inlining-once" safe_inlining_once

let safe_inlining = inlining_until_fixpoint safe_inlining_criterion

let () = String.Hashtbl.add opt_tbl "safe-inlining" safe_inlining

let eager_inlining_criterion opt x s list =
  Some (fun _ _ -> true)
  
let eager_inlining = inlining_once eager_inlining_criterion

let () = String.Hashtbl.add opt_tbl "eager-inlining" eager_inlining

let module_optimizations =
  ref [1, safe_inlining; 1, remove_unused]

let program_optimizations =
  ref []

let get_optim o =
  try String.Hashtbl.find opt_tbl o
  with Not_found -> Err.unknown_optimization o

let parse_optimizations desc =
  let make o =
    let coef, o =
      match String.split '*' o with
        | [o] -> (1, o)
        | [coef; o] -> (int_of_string coef, o)
        | _ -> Err.unknown_optimization o in
    coef, get_optim o in
  if desc = "" then []
  else List.map make (String.split ',' desc)

let set_module_optimizations desc =
  module_optimizations := parse_optimizations desc

let set_program_optimizations desc =
  program_optimizations := parse_optimizations desc  

let () =
  Cmdline.register ~long:"module-optimizations"
    (Cmdline.One set_module_optimizations)
    ~help:"set module-level optimizations list"

let () =
  Cmdline.register ~long:"program-optimizations"
    (Cmdline.One set_program_optimizations)
    ~help:"set program-level optimizations list"

let optimize_module opt =
  List.fold_left (fun opt (c, f) -> Int.iter f c opt) opt !module_optimizations

let optimize_program opt =
  List.fold_left (fun opt (c, f) -> Int.iter f c opt) opt !program_optimizations

